from __future__ import annotations

from asyncio import TaskGroup, sleep

from tests.conftest import SKIPIF_CI_AND_NOT_LINUX
from tests.test_redis import yield_test_redis
from utilities.pottery import yield_locked_resource
from utilities.text import unique_str
from utilities.timer import Timer


async def _coroutine(key: str) -> None:
    async with yield_test_redis() as redis, yield_locked_resource(redis, key):
        await sleep(0.1)


async def _asyncio_runner(num_tasks: int, key: str) -> None:
    async with TaskGroup() as tg:
        _ = [tg.create_task(_coroutine(key)) for _ in range(num_tasks)]


class TestYieldLockedResource:
    @SKIPIF_CI_AND_NOT_LINUX
    async def test_single_process(self) -> None:
        with Timer() as timer:
            await _asyncio_runner(3, unique_str())
        min_time = 0.3
        assert min_time <= float(timer) <= 3 * min_time



from __future__ import annotations

import asyncio
from contextlib import asynccontextmanager, suppress
from typing import TYPE_CHECKING

from pottery import AIORedlock
from pottery.exceptions import ReleaseUnlockedLock
from redis.asyncio import Redis

from utilities.datetime import MILLISECOND, SECOND, datetime_duration_to_float
from utilities.iterables import always_iterable

if TYPE_CHECKING:
    from collections.abc import AsyncIterator

    from utilities.types import Duration, MaybeIterable


@asynccontextmanager
async def yield_locked_resource(
    redis: MaybeIterable[Redis],
    key: str,
    /,
    *,
    duration: Duration = 10 * SECOND,
    sleep: Duration = MILLISECOND,
) -> AsyncIterator[None]:
    """Yield a locked resource."""
    masters = (  # skipif-ci-and-not-linux
        {redis} if isinstance(redis, Redis) else set(always_iterable(redis))
    )
    duration_use = datetime_duration_to_float(duration)  # skipif-ci-and-not-linux
    lock = AIORedlock(  # skipif-ci-and-not-linux
        key=key,
        masters=masters,
        auto_release_time=duration_use,
        context_manager_timeout=duration_use,
    )
    sleep_use = datetime_duration_to_float(sleep)  # skipif-ci-and-not-linux
    while not await lock.acquire():  # pragma: no cover
        _ = await asyncio.sleep(sleep_use)
    try:  # skipif-ci-and-not-linux
        yield
    finally:  # skipif-ci-and-not-linux
        with suppress(ReleaseUnlockedLock):
            await lock.release()


__all__ = ["yield_locked_resource"]
